(* 
Cardigan Compiler
    Reads in a Cardigan source file and compiles it
    to a java source file.
*)

open Printf
open Ast
open Hashtbl


module StringHash = Hashtbl.Make(struct
    type t = string
    let equal x y = x = y
    let hash = Hashtbl.hash
end)

(* ******************************************************************
    type: structRecord
    These are the values in our hash table. There are two valid forms.
        variable record: type * scopeLevel 
        function record: argument type list * return type 
    **************************************************************** *)
type structRecord =  
    Vrec of Ast.javaType * int
|   Frec of Ast.javaType list * Ast.javaType

(* ******************************************************************
    function: string_of_jtype
   
    Input: [Ast.javaType]
    Output: a string representing the Java type [string]
    **************************************************************** *)
let string_of_jtype = function
    JavaVoid    -> "void"
|   JavaBool    -> "boolean"
|   JavaInt     -> "int"
|   JavaFloat   -> "double"
|   JavaString  -> "String"
|   JavaInvalidType -> raise (Failure "Invalid Java type")
|   JavaCollection -> "cCollection"
|   JavaEnum  -> "cEnum"
|   JavaEnumItem -> "enum item"
|   JavaStruct  -> "cStruct"
|   Indeterminate -> "indeterminate"

let rec string_of_jtype_list  = function
    []      -> ""
|   a::[]   -> string_of_jtype a
|   a::b    -> (string_of_jtype a) ^ ", " ^ (string_of_jtype_list b)

(* ******************************************************************
    function: addVar
    Add a variable to the symbol table.
        If you try to modify a global variable, throw an error.
        If the variable is already there with the same type, proceed.
        If the variable is already there with indeterminate type, replace with this type.
        If the variable is already there with a different type, throw an error.
        If you try to add a function, thow an error (See addFunc, below.)

    Input: variable name [string], variable type [Ast.javaType], symbol table [StringHash], current scope level [int]
    Output: [unit]
    **************************************************************** *)
let addVar = fun key varType symbols varLevel ->
    if (StringHash.mem symbols key) then 
        match (StringHash.find symbols key) with
            Vrec(_, 0) -> 
                if varLevel == 0 then 
                    raise (Failure("Attempt to modify a global variable"))
                else
                    StringHash.add symbols key (Vrec(varType, varLevel))  
|           Vrec(currentType, currentLevel) -> (
                if currentType == varType then 
                    ()   
                else if currentType = Indeterminate then 
                    StringHash.replace symbols key (Vrec(varType, currentLevel))
                else
                    raise (Failure ("Attempt to assign " ^ (string_of_jtype varType) ^ " to " ^ (string_of_jtype currentType))))
|           Frec (_, _) -> raise (Failure ("Attempt to assign to function"))
    else
        StringHash.add symbols key (Vrec(varType, varLevel))

(* ******************************************************************
    function: getVarType
   
    Check the symbol table for the type of this variable.
        If the id is undefined, throw an error.
        If the id represents a function, throw an error.

    Input: variable name [string], symbol table [StringHash]
    Output: type of variable in table [Ast.javaType]
    **************************************************************** *)
let getVarType = fun key symbols->
    if (StringHash.mem symbols key) then match (StringHash.find symbols key) with
        Vrec(varType, _) -> varType
|       Frec(_, _) -> raise (Failure (key ^ " is a function"))
    else
        raise (Failure (key ^ " is undefined"))

(* ******************************************************************
    function: verifyType
   
    Check the symbol table to see if this type is allowed for this variable.
        If the variable is already there with the same type, proceed.
        If the variable is already there with indeterminate type, 
            replace with this type.
        If the variable is already there with a different type, throw an error.
        If the variable is not in the table, throw an error.

    Input: variable name [string], desired type [Ast.javaType], symbol table [StringHash]
    Output: [unit]
    **************************************************************** *)
let verifyType = fun key desiredType symbols ->
    if (StringHash.mem symbols key) then match (StringHash.find symbols key) with
        Vrec(varType, currentLevel) -> (
            if varType == desiredType then 
                ()
            else if varType = Indeterminate then 
                StringHash.replace symbols key (Vrec(varType, currentLevel))
            else 
                raise (Failure((string_of_jtype varType) ^ " " ^ key ^ " is being treated as " ^ (string_of_jtype desiredType)))) 
|       Frec(_, _) -> raise (Failure (key ^ " is a function"))
    else
        raise (Failure (key ^ " is undefined"))

(* ******************************************************************
    function: clearScope
    and eval
    Clear all variables from a certain scope level out of the symbol table.
    Call this when done with that scope.
    
    Input: scope to clear [int], symbol table [StringHash]
    Output: [unit]
    **************************************************************** *)
let clearScope = fun scope symbols ->
    let cleanup key = function
        Vrec(varType, varScope) -> if varType == Indeterminate then
                                        raise (Failure("The type of " ^ key ^ " is indeterminate"))
                                   else if varScope == scope then 
                                        StringHash.remove symbols key
                                  else ()
|       Frec(_, _) -> ()
    in
    StringHash.iter cleanup symbols

(* ******************************************************************
    function: checkFuncName
    
    Check if a certain function name is defined.
    
    Input: function name [string], symbol table [StringHash]
    Output: [boolean]
    **************************************************************** *)
let checkFuncName = fun id symbols ->
    StringHash.mem symbols id

(* ******************************************************************
    function: callFunction
    
    Call a function that is already defined in the symbol table.
    
    Input: function name [string], symbol table [StringHash]
    Output: object containing argument types and return types
            [Frec(Ast.javaType list, Ast.javaType)]
    **************************************************************** *)
let callFunction = fun id symbols ->
    if (StringHash.mem symbols id) then
        StringHash.find symbols id
    else
        raise (Failure("Called undefined function -- This shouldn't happen because you can use addRecFunc to call a function you haven't seen yet"))
(* ******************************************************************
    function: addFunc
    
    Add a new function to the symbol table.
        For recursive functions, see addRecFunc, below.
        If function is already defined, throw error.

    Input: function name [string], list of argument types [Ast.javaType list],
            return type [Ast.javaType], symbol table [StringHash]
    Output: [unit]
    **************************************************************** *)
let addFunc = fun id arguments funcType symbols ->
    if (StringHash.mem symbols id) then
        raise (Failure("Trying to redefine previously defined function"))
    else
        StringHash.add symbols id (Frec(arguments, funcType))

(* ******************************************************************
    function: addRecFunc
    
    Add a new recursive function to the symbol table.
    Sets return type to indeterminate.
    
    Input: function name [string], list of argument types [Ast.javaType list],
            symbol table [StringHash]
    Output: [unit]
    **************************************************************** *)
let addRecFunc = fun id arguments symbols ->
    if (StringHash.mem symbols id) then
        raise (Failure("Trying to define a recursive function that was already defined -- this shouldn't happen because you should have used checkFuncName first"))
    else
        StringHash.add symbols id (Frec(arguments, Indeterminate))

(* ******************************************************************
    function: setFuncType
    
    Set the return type of a previously defined function.
        If id is in the table as a variable, throw an error.
        If id is not in the table, throw an error.

    Input: function name [string], return type [Ast.javaType],
            symbol table [StringHash]
    Output: [unit]
    **************************************************************** *)
let setFuncType = fun id retType symbols ->
    if (StringHash.mem symbols id) then match (StringHash.find symbols id) with
        Frec(args, _) -> StringHash.replace symbols id (Frec(args, retType))
|       Vrec(_, _) -> raise (Failure ("Trying to define the return type of a variable -- no such thing"))
    else 
        raise (Failure ("Trying to define the return type of a non-existent function. You probably should use addFunc instead"))

(* ******************************************************************
    function: getFuncType
    
    Get the return type of a previously defined function.
        If id is not in the table as a function, throw an error.
        If id is not in the table, throw an error.

    Input: function name [string],
            symbol table [StringHash]
    Output: [Ast.javaType]
    **************************************************************** *)
let getFuncType = fun id symbols ->
    if (checkFuncName id symbols) then (match (StringHash.find symbols id) with
        Frec(_,retType) -> retType
        |   _ -> raise (Failure "Not a valid function"))
    else raise (Failure ("Undefined function: " ^ id))

(* ******************************************************************
    function: string_of_identifier
    
    Input: [Ast.identifier]
    Output: [string]
    **************************************************************** *)
let string_of_identifier = function
    Identifier(a) -> a

(* ******************************************************************
    function: check_for_play
    
    Replace the function name with "main"
    If the function name is PLAY
    
    Input: [string]
    Output: [string]
    **************************************************************** *)
let check_for_play = function
    "PLAY"  -> "main"
|   x       -> x

(* ******************************************************************
    function: jtype_of_string
    
    Get relevant java type of the input string
    If the input string is an invalid java type, throw an error
   
    Input: [string]
    Output: [Ast.javaType]
    **************************************************************** *)
let jtype_of_string x = match x with
    "void"      -> JavaVoid
|   "boolean"   -> JavaBool
|   "int"       -> JavaInt
|   "double"     -> JavaFloat
|   "String"    -> JavaString
|   _           -> raise (Failure( "No corresponding Java type for " ^ x))
   
let rec concat (sList : string list) : string =
  match sList with
    [] -> ""
  | a :: b -> a ^ (concat b) ^ ";\n"

let prepend_javatype t str = 
    (string_of_jtype t) ^ " " ^ str

let rec eval_arglist lst = match lst with
    []  -> ""
|   Identifier(a)::[] -> a
|   Identifier(a)::b  -> a^", "^(eval_arglist b)

(* ******************************************************************
    function: get_collection_as_list
    
    Return a list for elements in a collection
    If the input expression is not a valid collection, throw an error
   
    Input: [Ast.expression]
    Output: a list of [Ast.expression]
    **************************************************************** *)
let get_collection_as_list = function
    Collection(x)          -> x
    (*
|   CollectionBounds(x,y)  -> "collectionbound" *)
|   _                      -> raise (Failure "Invalid collection.")

(* ******************************************************************
    function: get_enum_as_list
    
    Return a list for elements in an enum
    If the input expression is not a valid enum, throw an error
   
    Input: [Ast.expression]
    Output: a list of [Ast.expression]
    **************************************************************** *)
let get_enum_as_list = function
    Enum(x)         -> x
|   _               -> raise (Failure "Invalid enum.")
(* ******************************************************************
    function: get_struct_as_list
    
    Return a list for elements in a struct
    If the input expression is not a valid struct, throw an error
   
    Input: [Ast.expression]
    Output: a list of [Ast.expression]
    **************************************************************** *)
let get_struct_as_list = function
    Struct(x)       -> x
|   _               -> raise (Failure "Invalid struct.")

(* ******************************************************************
    function: type_of_struct_record
   
    Input: [structRecord]
    Output: [Ast.javaType]
    **************************************************************** *)
let type_of_struct_record = function
    Vrec(t,level) -> t
|   Frec(argst,t) -> t

(* ******************************************************************
    function: eval_binary_op
    
    Get the result of binary operation as a string
    If the type of expression is invalid, throw an error
   
    Input: [Ast.expression]
    Output: [string]
    **************************************************************** *)
let rec eval_binary_op expr currentFunction funcArgs symbols nestingLevel = match expr with
    Plus(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
                JavaInt, JavaInt -> x1 ^"+"^ y1
            |   JavaInt, JavaFloat -> x1 ^"+"^ y1
            |   JavaFloat, JavaInt -> x1 ^"+"^ y1
            |   JavaFloat, JavaFloat -> x1 ^"+"^ y1
            |   JavaString, JavaString -> x1 ^ "+" ^ y1
            |   _,_ -> raise (Failure ("Unable to evaluate plus on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))
        )
|   Minus(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
                JavaInt, JavaInt -> x1 ^"-"^ y1
            |   JavaInt, JavaFloat -> x1 ^"-"^ y1
            |   JavaFloat, JavaInt -> x1 ^"-"^ y1
            |   JavaFloat, JavaFloat -> x1 ^"-"^ y1
            |   _,_ -> raise (Failure ("Unable to evaluate minus on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))        )
|   Times(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
                JavaInt, JavaInt -> x1 ^"*"^ y1
            |   JavaInt, JavaFloat ->  x1 ^"*"^ y1
            |   JavaFloat, JavaInt -> x1 ^"*"^ y1
            |   JavaFloat, JavaFloat ->  x1 ^"*"^ y1
            |   _,_ -> raise (Failure ("Unable to evaluate times on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))        )
|   Divide(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
                JavaInt, JavaInt -> x1 ^"/"^ y1
            |   JavaInt, JavaFloat -> x1 ^"/"^ y1
            |   JavaFloat, JavaInt -> x1 ^"/"^ y1
            |   JavaFloat, JavaFloat -> x1 ^"/"^ y1
            |   _,_ -> raise (Failure ("Unable to evaluate divide on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))        )
|   Mod(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
                JavaInt, JavaInt -> x1 ^ "%" ^ y1
            |   _,_ -> raise (Failure ("Unable to evaluate mod on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))        )
|   And(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
                JavaBool, JavaBool -> x1 ^ "&&" ^ y1
            |   _,_ -> raise (Failure ("Unable to evaluate AND on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))        )
|   Or(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
                JavaBool, JavaBool -> x1 ^ "||" ^ y1
            |   _,_ -> raise (Failure ("Unable to evaluate OR on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))        )
|   Lt(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
            JavaInt, JavaInt -> x1 ^ "<" ^ y1  
        |   JavaInt, JavaFloat -> x1 ^ "<" ^ y1
        |   JavaFloat, JavaInt -> x1 ^ "<" ^ y1
        |   JavaFloat, JavaFloat -> x1 ^ "<" ^ y1
        |   _,_ -> raise (Failure ("Unable to evaluate less than on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))    )
|   Gt(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
            JavaInt, JavaInt -> x1 ^ ">" ^ y1
        |   JavaInt, JavaFloat -> x1 ^ ">" ^ y1
        |   JavaFloat, JavaInt -> x1 ^ ">" ^ y1
        |   JavaFloat, JavaFloat -> x1 ^ ">" ^ y1
        |   _,_ -> raise (Failure ("Unable to evaluate greater than on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))    )
|   Geq(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
            JavaInt, JavaInt -> x1 ^ ">=" ^ y1
        |   JavaInt, JavaFloat -> x1 ^ ">=" ^ y1
        |   JavaFloat, JavaInt -> x1 ^ ">=" ^ y1
        |   JavaFloat, JavaFloat -> x1 ^ ">=" ^ y1
        |   _,_ -> raise (Failure ("Unable to evaluate greater than/equal on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))    )
|   Leq(x,y) -> (
            let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
            let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in 
            let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
            let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
            match (xt,yt) with
            JavaInt, JavaInt -> x1 ^ "<=" ^ y1 
        |   JavaInt, JavaFloat -> x1 ^ "<=" ^ y1
        |   JavaFloat, JavaInt -> x1 ^ "<=" ^ y1
        |   JavaFloat, JavaFloat -> x1 ^ "<=" ^ y1
        |   _,_ -> raise (Failure ("Unable to evaluate less than/equal on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))    )
    
(*TODO: handle collections, structs, enums, variables better in neq, eq *)
|   Neq(x,y) -> (
        let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
        let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in
        x1 ^ "!=" ^ y1
    )
|   Eq(x,y) -> (
        let x1 = eval_expr x currentFunction funcArgs symbols nestingLevel in 
        let y1 = eval_expr y currentFunction funcArgs symbols nestingLevel in
        x1 ^ "==" ^ y1
    )
(* ******************************************************************
    function: type_of_binary_op
    
    Return the result of binary operation
    If the type of expression is invalid, throw an error
   
    Input: [Ast.expression]
    Output: [Ast.javaType]
    **************************************************************** *)
and type_of_binary_op expr currentFunction funcArgs symbols nestingLevel = match expr with
    Plus(x,y) -> (
        let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
        let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
        match (xt,yt) with
            JavaInt, JavaInt -> JavaInt
|           JavaInt, JavaFloat -> JavaFloat
|           JavaFloat, JavaInt -> JavaFloat
|           JavaFloat, JavaFloat -> JavaFloat
|           JavaString, JavaString -> JavaString
 |   _,_ -> raise (Failure ("Unable to evaluate binary operation on " ^ "(" ^ (string_of_jtype xt) ^ ", " ^ (string_of_jtype yt)^")"))
    )
|   Minus(x,y) -> (
        let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
        let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
        match (xt,yt) with
            JavaInt, JavaInt -> JavaInt
|           JavaInt, JavaFloat -> JavaFloat
|           JavaFloat, JavaInt -> JavaFloat
|           JavaFloat, JavaFloat -> JavaFloat
|           _, _ -> raise (Failure ("Unable to evaluate binary operation"))
    )
|   Times(x,y) -> (
        let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
        let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
        match (xt,yt) with
            JavaInt, JavaInt -> JavaInt
|           JavaInt, JavaFloat -> JavaFloat
|           JavaFloat, JavaInt -> JavaFloat
|           JavaFloat, JavaFloat -> JavaFloat
|           _, _ -> raise (Failure ("Unable to evaluate binary operation"))
    )
|   Divide(x,y) -> (
        let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
        let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
        match (xt,yt) with
            JavaInt, JavaInt -> JavaInt
|           JavaInt, JavaFloat -> JavaFloat
|           JavaFloat, JavaInt -> JavaFloat
|           JavaFloat, JavaFloat -> JavaFloat
|           _, _ -> raise (Failure ("Unable to evaluate binary operation"))
    )
|   Mod(x,y) -> (
        let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
        let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
        match (xt,yt) with
            JavaInt, JavaInt -> JavaInt
|           _, _ -> raise (Failure ("Unable to evaluate binary operation"))
    )
|   And(x,y) -> (
        let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
        let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
        match (xt,yt) with
            JavaBool, JavaBool -> JavaBool
|           _, _ -> raise (Failure ("Unable to evaluate binary operation"))
    )
|   Or(x,y) -> (
        let xt = type_of_expr x currentFunction funcArgs symbols nestingLevel in
        let yt = type_of_expr y currentFunction funcArgs symbols nestingLevel in
        match (xt,yt) with
            JavaBool, JavaBool -> JavaBool
|           _, _ -> raise (Failure ("Unable to evaluate binary operation"))
    )
|   Neq(x,y) -> JavaBool
|   Eq(x,y) -> JavaBool
|   Lt(x,y) -> JavaBool
|   Gt(x,y) -> JavaBool
|   Leq(x,y) -> JavaBool
|   Geq(x,y) -> JavaBool

(* ******************************************************************
    function: cast
    
    Return the result of cast
    If the type of expression is invalid, throw an error
   
    Input: input [Ast.expression], 
           input type [Ast.javaType], output type [Ast.javaType],
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash]
           nesting level [int]
    Output: a list of cast result and its type [string, Ast.javaType]
    **************************************************************** *)
and cast input inputtype outputtype currentFunction funcArgs symbols nestingLevel = 
    match inputtype with
    JavaBool    -> (cast_bool input outputtype currentFunction funcArgs symbols nestingLevel, jtype_of_string outputtype)
|   JavaInt     -> (cast_int input outputtype currentFunction funcArgs symbols nestingLevel, jtype_of_string outputtype)
|   JavaFloat   -> (cast_float input outputtype currentFunction funcArgs symbols nestingLevel, jtype_of_string outputtype)
|   JavaString  -> (cast_string input outputtype currentFunction funcArgs symbols nestingLevel, jtype_of_string outputtype)
|   _           -> raise (Failure ("Invalid cast from " ^ (string_of_jtype inputtype)))
(* ******************************************************************
    function: cast_bool, cast_int, cast_float, cast_string
    
    Cast bool (int/ float/ string) input to another type
    If the type of expression is invalid, throw an error
   
    Input: input [Ast.expression], 
           output type [Ast.javaType],
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash]
           nesting level [int]
    Output: a list of cast result and its type [string, Ast.javaType]
    **************************************************************** *)  
and cast_bool input outputtype currentFunction funcArgs symbols nestingLevel = 
    try (let value = bool_of_string (eval_expr input currentFunction funcArgs symbols nestingLevel) in
    match outputtype with 
    "boolean"   -> string_of_bool value
|   "int"       -> (match value with 
                        true    -> "1"
                    |   false   -> "0")
|   "double"     -> (match value with 
                        true    -> "1.0"
                    |   false   -> "0.0")
|   "String"    -> (match value with 
                        true    -> "\"1\""
                    |   false   -> "\"0\"")
|   _           -> raise (Failure ("Invalid cast from boolean to " ^ outputtype))
) with Failure(x) -> raise (Failure ("Invalid cast from boolean to " ^ outputtype))

and cast_int input outputtype currentFunction funcArgs symbols nestingLevel = (try (let value = int_of_string (eval_expr input currentFunction funcArgs symbols nestingLevel) in
    match outputtype with
    "boolean"   -> (match value with 
                        0       -> "false"
                    |   _       -> "true" )
|   "int"       -> string_of_int value
|   "double"     -> (string_of_int value) ^ ".0"
|   "String"    -> "\"" ^ (string_of_int value) ^ "\""
|   _           -> raise (Failure ("Invalid cast from int to " ^ outputtype))
) with Failure(x) -> raise (Failure ("Invalid cast from int to " ^ outputtype)))

and cast_float input outputtype currentFunction funcArgs symbols nestingLevel = 
    (try (let value = float_of_string (eval_expr input currentFunction funcArgs symbols nestingLevel) in match outputtype with
        "boolean" -> (match value with 
            0.0       -> "false"
|           _         -> "true"
        )
|       "int" -> (let p = String.index (string_of_float value) '.' in 
                 String.sub (string_of_float value) 0 p)
|       "double" -> string_of_float value
|       "String" -> "\"" ^ (string_of_float value) ^ "\""
|       _  -> raise (Failure ("Invalid cast from float to " ^ outputtype))
    ) with Failure(x) -> raise (Failure ("Invalid cast from float to " ^ outputtype)))


and cast_string input outputtype currentFunction funcArgs symbols nestingLevel = 
    (try (let value = eval_expr input currentFunction funcArgs symbols nestingLevel in match outputtype with
        "boolean" -> (match value with 
            "" -> "false"
|           _  -> "true"
        )
|       "int" -> (let v = String.sub value 1 ((String.length value)-2) in 
                try (string_of_int (int_of_string v)) 
                with Failure(a) -> raise (Failure "Invalid cast from string to int.")
            )
|       "double" -> 
                (let v = String.sub value 1 ((String.length value)-2) in 
                try (string_of_float (float_of_string v)) 
                with Failure(a) -> raise (Failure "Invalid cast from string to float."))
|       "String" -> value
|       _ -> raise (Failure ("Invalid cast from string to " ^ outputtype))
    ) with Failure(x) -> raise (Failure ("Invalid cast from string to " ^ outputtype)))
(* ******************************************************************
    function: eval_expr
    
    Return the value of expression as a java-valid string
  
    Input: input [Ast.expression], 
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash]
           nesting level [int]
    Output: [string]
    **************************************************************** *)  
and eval_expr expr currentFunction funcArgs symbols nestingLevel = match expr with
      (* Evaluate expressions by converting them to their java representations *) 
    Integer(x) -> string_of_int x
|   Floating(x) -> string_of_float x
|   Boolean(x) -> string_of_bool x
|   StringString(x) -> x
|   IdExpr(Identifier(x)) -> x
|   Collection(x) -> "["^ (eval_expr_list x currentFunction funcArgs symbols nestingLevel) ^ "]"
|   CollectionBounds(x,y) -> "["^
                             (eval_expr x currentFunction funcArgs symbols nestingLevel)^
                             ":"^
                             (eval_expr y currentFunction funcArgs symbols nestingLevel)^
                             "]"
|   ParenExpr(x) -> "(" ^ 
                    (eval_expr x currentFunction funcArgs symbols nestingLevel) ^
                    ")"
|   Enum(x) -> "|" ^ eval_arglist x ^ "|"
|   Cast(x, y) -> fst (cast x (type_of_expr x currentFunction funcArgs symbols nestingLevel)
                      y currentFunction funcArgs symbols nestingLevel)
|   Input(x) -> "nextLine();"
|   Negate(x) -> (match (type_of_expr x currentFunction funcArgs symbols nestingLevel) with
                      JavaInt -> (let y = (match x with 
                                  Integer(a) -> string_of_int a
                            | IdExpr(id) -> string_of_identifier id
                            | _ -> raise (Failure ("Cannot invert a value of type " ^
                                          (string_of_jtype (type_of_expr x currentFunction funcArgs symbols nestingLevel)))))
                            in "(-" ^ y ^ ")")
                |   JavaFloat -> (let y = (match x with 
                                  Floating(a) -> string_of_float a
                              | IdExpr(id) -> string_of_identifier id
                              | _ -> raise (Failure ("Cannot invert a value of type " ^
                                            (string_of_jtype (type_of_expr x currentFunction funcArgs symbols nestingLevel)))))
                            in "(-" ^ y ^ ")")
                | _ -> raise (Failure ("Cannot invert a value of type " ^
                                      (string_of_jtype (type_of_expr x currentFunction funcArgs symbols nestingLevel)))))
|   Invert(x) -> (match (type_of_expr x currentFunction funcArgs symbols nestingLevel) with
                  JavaBool -> (let y = (match x with 
                               Boolean(true) -> "true"
                           | Boolean(false) -> "false"
                           | IdExpr(id) -> string_of_identifier id
                           | _ -> raise (Failure ("Cannot invert a value of type " ^
                                                 (string_of_jtype (type_of_expr x currentFunction funcArgs symbols nestingLevel))))) in "(!" ^ y ^ ")")
              | _ -> raise (Failure ("Cannot invert a value of type " ^ (string_of_jtype (type_of_expr x currentFunction funcArgs symbols nestingLevel)))))
|   BinaryOp(x) -> (eval_binary_op x currentFunction funcArgs symbols nestingLevel)
|   Increment(Integer(x)) -> "("^(string_of_int x) ^ "+1)"
|   Increment(Floating(x))-> "("^(string_of_float x) ^ "+1.0)"
|   Increment(IdExpr(Identifier(x))) -> if (getVarType x symbols)==JavaInt then 
                                           x^"++;"
                                        else if (getVarType x symbols)==JavaFloat then
                                           x^"+1.0;"
                                        else raise (Failure ("Invalid type"^(string_of_jtype (getVarType x symbols))))
|   Increment(x)  -> raise (Failure ("Cannot increment a value of type "))
|   Decrement(Integer(x)) -> "("^(string_of_int x) ^ "-1)"
|   Decrement(Floating(x))-> "("^(string_of_float x) ^ "-1.0)"
|   Decrement(IdExpr(Identifier(x))) -> if (getVarType x symbols)==JavaInt then 
                                           "("^x^"-1)"
                                        else if (getVarType x symbols)==JavaFloat then
                                           "("^x^"-1.0)"
                                        else raise (Failure ("Cannot increment a value of type "^(string_of_jtype (getVarType x symbols))))
|   Decrement(x)  -> raise (Failure ("Cannot decrement a value of type "))
|   CollectionIndex(Identifier(x),Integer(y)) -> 
        x ^ "[" ^ string_of_int y ^"]"
|   CollectionIndex(Identifier(x),y) -> x ^ "[" ^ (eval_expr y currentFunction funcArgs symbols nestingLevel)^"]"
|   Lookup(Identifier(x),y) -> let exprstr = x ^"."^(eval_expr y currentFunction funcArgs symbols nestingLevel) in
                               if (getVarType x symbols)==JavaCollection then
                                  exprstr^";"
                               else if (getVarType x symbols)==JavaEnum && (StringHash.mem symbols exprstr) then
                                  "globalEnum.find(\""^exprstr^"\").enumValue"
                               else
                                  raise (Failure ("Cannot find: "^(eval_expr y currentFunction funcArgs symbols nestingLevel)^" in enum "^x))          
|   FunctionCall(FunctionIdentifier(a, b)) -> 
    (* TODO: check argument types *)
    (check_for_play a) ^ "("^eval_arglist b ^")"
|   Struct(x) -> "new cStruct();\n"^ 
                 (eval_structdef_list x currentFunction funcArgs symbols nestingLevel)

(* ******************************************************************
    function: eval_expr_list
    
    Look through a list of expressions and evaluate
    each separately, returning a comma separated list
    of the result
    
    Input: a list of [Ast.expression], 
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash]
           nesting level [int]
    Output: [string]
    **************************************************************** *)  
and eval_expr_list lst currentFunction funcArgs symbols nestingLevel= match lst with
    []    -> ""
|   a::[] -> eval_expr a currentFunction funcArgs symbols nestingLevel
|   a::b -> (eval_expr a currentFunction funcArgs symbols nestingLevel)^", "^
    (eval_expr_list b currentFunction funcArgs symbols nestingLevel)
    
(* ******************************************************************
    function: eval_structdef_list
    
    Look through a list of expressions and evaluate
    each separately, returning a java-valid list
    of the result
    
    Input: a list of [Ast.expression], 
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash]
           nesting level [int]
    Output: [string]
    **************************************************************** *) 
and eval_structdef_list lst currentFunction funcArgs symbols nestingLevel = match lst with
    []    -> ""
|   StructDef(s,e)::[]  -> "\n"^(String.make nestingLevel '\t')^string_of_jtype (type_of_expr e currentFunction funcArgs symbols nestingLevel)^
                           " "^
                           s^
                           "="^
                           (eval_expr e currentFunction funcArgs symbols nestingLevel)^
                           ";"
|   StructDef(s,e)::l   -> "\n"^(String.make nestingLevel '\t')^string_of_jtype (type_of_expr e currentFunction funcArgs symbols nestingLevel)^
                           " "^
                           s^
                           "="^
                           (eval_expr e currentFunction funcArgs symbols nestingLevel)^
                           ";"^ 
                           eval_structdef_list l currentFunction funcArgs symbols nestingLevel

(* ******************************************************************
    function: string_of_struct_record
    
    Get a string which contains the type and nesting level
    of a record in the symbol
    
    Input: a record [StructRecord], 
           symbols [StingHash]
           nesting level [int]
    Output: [string]
    **************************************************************** *) 
and string_of_struct_record r symbols nestingLevel = match r with
    Vrec(t,level) -> (string_of_jtype t) ^" " ^(string_of_int level)
|   Frec(tlist,t) -> (string_of_jtype t)^": "^(string_of_jtype_list tlist)

(* ******************************************************************
    function: print_hash_record
    
    Print type and nesting level of each elements in the symbol table
    
    Input: a record [StructRecord], 
           symbols [StingHash]
           nesting level [int]
    Output: [string]
    **************************************************************** *)
and print_hash_record symbols nestingLevel key value  =
    print_endline ("\t"^key ^ " "^(string_of_struct_record value symbols nestingLevel))
    
(* ******************************************************************
    function: print_hash
    
    Print symbol table
    
    Input: symbols [StingHash],
           nesting level [int]
    Output: [string]
    **************************************************************** *)
and print_hash symbols nestingLevel = 
    ignore (print_endline "Contents of symbol table: ");
    ignore (StringHash.iter (print_hash_record symbols nestingLevel) symbols);
    print_endline ""

(* ******************************************************************
    function: save_assignment_to_hash
    
    Save assignment to symbol table, in which
    each element in the symbol table has type of StructType
    Print symbol table 
        
    Input: identifier [Ast.identifier],
           expression [Ast.expression],
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash],
           nesting level [int]
    Output: [string]
    **************************************************************** *)
and save_assignment_to_hash idstr expr currentFunction funcArgs symbols nestingLevel = 
    (* TODO check if idstr is already in symbols *)
    ignore (addVar idstr (type_of_expr expr
        currentFunction funcArgs symbols nestingLevel)
        symbols nestingLevel);
    ignore (if (List.mem (Identifier(idstr)) funcArgs) then 
              (List.iteri (fun i a -> match a with 
              Identifier(idstr)   -> (add_argtype i idstr currentFunction funcArgs symbols nestingLevel)) funcArgs)
            else ());
    match expr with
    Collection(lst) -> ignore (save_collection_elements lst idstr 0 currentFunction funcArgs symbols nestingLevel); 
        print_hash symbols nestingLevel
|   CollectionBounds(x,y) -> 
        let low = int_of_string (eval_expr x currentFunction funcArgs symbols nestingLevel) in
        let high = int_of_string (eval_expr y currentFunction funcArgs symbols nestingLevel) in
        ignore (save_collectionbounds_element low high idstr symbols nestingLevel);
        print_hash symbols nestingLevel 
|   Enum(lst) -> ignore (save_enum_elements lst idstr symbols nestingLevel);
        print_hash symbols nestingLevel
|   Struct(lst) ->ignore(save_struct_elements lst idstr currentFunction funcArgs symbols nestingLevel);
        print_hash symbols nestingLevel
|   _ -> print_hash symbols nestingLevel


and match_arg_types b argTypes symbols nestingLevel = 
    ()

(* ******************************************************************
    function: type_of_expr
    
    Return type of expression
        
    Input: expression [Ast.expression],
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash],
           nesting level [int]
    Output: [Ast.javaType]
    **************************************************************** *)
and type_of_expr ex currentFunction funcArgs symbols nestingLevel= match ex with
    Integer(x) -> JavaInt
|   Boolean(x) -> JavaBool
|   Floating(x) -> JavaFloat
|   StringString(x) -> JavaString
|   ParenExpr(x) -> type_of_expr x currentFunction funcArgs symbols nestingLevel
|   IdExpr(Identifier(x)) -> getVarType x symbols
|   Collection(x) -> JavaCollection
|   CollectionIndex(Identifier(x),Integer(y)) -> (try (
        let idstr = (x^"["^(string_of_int y)^"]") in
             if (StringHash.mem symbols idstr) then
               getVarType idstr symbols
             else
               raise (Failure ("Invalid index "^(string_of_int y))))
        with Not_found -> raise (Failure "Failed to access collection"))
|   CollectionIndex(Identifier(x), IdExpr(Identifier(y))) -> 
              if (getVarType y symbols)==JavaInt then
                  Indeterminate
              else
                  raise (Failure ("Invalid index "^y))
|   CollectionBounds(Integer(x),Integer(y)) -> if (x<y) then JavaCollection
           else
             raise (Failure ("Invalid Bound from "^
                            (string_of_int x)^
                            " to "^
                            (string_of_int y)))
|   Enum(x) -> JavaEnum
|   Cast(x,y) -> jtype_of_string y
|   Input(s) -> jtype_of_string s
|   Negate(Integer(x)) -> JavaInt
|   Negate(Floating(x)) -> JavaFloat
|   Negate(IdExpr(Identifier(x))) -> if ((getVarType x symbols)==JavaInt) 
    then JavaInt else (if (getVarType x symbols)==JavaFloat then JavaFloat
    else raise (Failure("Cannot negate a value of type " ^
        (string_of_jtype (getVarType x symbols )))))
|   Invert(Boolean(x)) -> JavaBool
|   Invert(IdExpr(Identifier(x))) -> if ((getVarType x symbols)==JavaBool)
    then JavaBool else raise (Failure ("Cannot invert a value of type " ^
        (string_of_jtype (getVarType x symbols ))))
|   BinaryOp(x) ->  type_of_binary_op x currentFunction funcArgs symbols nestingLevel
|   Increment(Integer(x)) -> JavaInt
|   Increment(Floating(x)) -> JavaFloat
|   Increment(IdExpr(Identifier(x))) -> if ((getVarType x symbols)==JavaInt)
    then JavaInt else (if (getVarType x symbols)==JavaFloat then JavaFloat
    else raise (Failure ("Cannot increment a value of type " ^
        (string_of_jtype (getVarType x symbols )))))
|   Decrement(Integer(x)) -> JavaInt
|   Decrement(Floating(x)) -> JavaFloat
|   Decrement(IdExpr(Identifier(x))) -> if ((getVarType x symbols)==JavaInt)
    then JavaInt else (if (getVarType x symbols)==JavaFloat then JavaFloat
    else raise (Failure ("Cannot decrement a value of type " ^
        (string_of_jtype (getVarType x symbols )))))
|   Lookup(Identifier(x), y) -> let t=(getVarType x symbols) in
      if (t==JavaEnum) then JavaInt else t
|   FunctionCall(FunctionIdentifier(a, b)) -> (try (
       match (callFunction a symbols) with 
            Frec(argTypes,retType) ->(
                ignore(match_arg_types b argTypes symbols nestingLevel);
                retType)
        |   _                       -> raise (Failure "Not a function") 
    ) with Failure(x) -> raise (Failure ("Undefined function "^ a)))


(* type_of_return_var x:: finds the type_of_expr of return var, if none then False *) 
|   Struct(x) -> JavaStruct
|   _         -> raise (Failure "Invalid type")

(* Note: This function uses a nested if/else statement rather than pattern matching. This is intentional - I require an equality match, which O'Caml does not provide. In general, I use matching over if/else in O'Caml. *)
and add_argtype index idstr currentFunction funcArgs symbols nestingLevel = 
    let newType = getVarType idstr symbols in
    let newFrec = try (match (StringHash.find symbols currentFunction) with 
            Frec(argTypes,ret) ->  Frec((List.mapi 
                        (fun i e -> 
                            if ((i == index) && ((e == Indeterminate) || (e == newType)))
                                then newType
                                else (if (i == index) then raise (Failure "Incompatible argument type") else e))
                     argTypes), ret)
            | _ -> raise (Failure "Not a function"))
    
        with Failure(x) -> raise (Failure 
            ("Cannot update arguments for undefined function "^currentFunction)) in
        (* ignore(print_endline (string_of_jtype newType));
        (ignore(print_endline (string_of_struct_record newFrec symbols nestingLevel)));
 *)
            StringHash.replace symbols currentFunction newFrec
(* ******************************************************************
    function: add_collection_element
    
    Save one collection element into symbol table
        
    Input: collection element [Ast.expression],
           identifier [string],
           index of element [int]
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash],
           nesting level [int]
    Output: [unit]
    **************************************************************** *)
and add_collection_element = fun el id ind currentFunction funcArgs symbols nestingLevel ->
    addVar
            (id ^ "[" ^ (string_of_int ind) ^ "]")
            (type_of_expr el currentFunction funcArgs symbols nestingLevel)
            symbols
            nestingLevel

(* ******************************************************************
    function: save_collection_elements
    
    Save collection elements into symbol table
        
    Input: a list of collection elements [Ast.expression],
           identifier [string],
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash],
           nesting level [int]
    Output: [unit]
    **************************************************************** *)
and save_collection_elements = fun x id start currentFunction funcArgs symbols nestingLevel -> 
    match x with
    []          -> ()
|   el::[]      -> add_collection_element el id start  currentFunction funcArgs symbols nestingLevel
|   el::l       -> ignore (add_collection_element el id start currentFunction funcArgs symbols nestingLevel);
            (save_collection_elements l id (start+1) currentFunction funcArgs symbols nestingLevel)
(* ******************************************************************
    function: add_struct_element
    
    Save one struct element into symbol table
        
    Input: identifier [string],
           property/function name [string],
           expression [Ast.expression],
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash],
           nesting level [int]
    Output: [unit]
    **************************************************************** *)
and add_struct_element outerstr innerstr expr currentFunction funcArgs symbols nestingLevel = 
    addVar (outerstr ^ "." ^ innerstr)
        (type_of_expr expr currentFunction funcArgs symbols nestingLevel)
        symbols nestingLevel
(* ******************************************************************
    function: save_struct_elements
    
    Save struct elements into symbol table
        
    Input: a list of struct elements [Ast.expression],
           identifier [string],
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash],
           nesting level [int]
    Output: [unit]
    **************************************************************** *)
and save_struct_elements lst outerstr currentFunction funcArgs symbols nestingLevel = match lst with
    []          -> ()
|   StructDef(innerstr,expr)::[]      -> 
        add_struct_element outerstr innerstr expr currentFunction funcArgs symbols nestingLevel
|   StructDef(innerstr,expr)::l       ->
        ignore (add_struct_element outerstr innerstr expr currentFunction funcArgs symbols nestingLevel);
        save_struct_elements l outerstr currentFunction funcArgs symbols nestingLevel

(* ******************************************************************
    function: save_collectionbounds_element
    
    Save elements of a collection with sequential integer numbers
    into symbol table
        
    Input: lower bound of element [int],
           higher bound of element [int],
           identifier [string],
           symbols [StingHash],
           nesting level [int]
    Output: [unit]
    **************************************************************** *)
and save_collectionbounds_element low high id symbols nestingLevel = 
    for i=low to high do
      addVar
            (id ^ "[" ^ (string_of_int (i-low)) ^ "]")
            JavaInt
            symbols
            nestingLevel
    done
    
(* ******************************************************************
    function: add_enum_element
    
    Save one enum element into symbol table
            
    Input: enum element [Ast.expression],
           identifier [string],
           symbols [StingHash],
           nesting level [int]
    Output: [unit]
    **************************************************************** *)
and add_enum_element = fun el idstr symbols nestingLevel -> match el with
    Identifier(inner) -> addVar (idstr ^ "." ^ inner) JavaEnumItem symbols nestingLevel

(* ******************************************************************
    function: save_enum_element
    
    Save enum elements into symbol table
            
    Input: a list of enum elements [Ast.expression],
           identifier [string],
           symbols [StingHash],
           nesting level [int]
    Output: [unit]
    **************************************************************** *)
and save_enum_elements = fun x id symbols nestingLevel ->
    match x with
        [] -> ()
|       el::[] -> add_enum_element el id symbols nestingLevel
|       el::l -> 
            ignore (add_enum_element el id symbols nestingLevel);
            (save_enum_elements l id symbols nestingLevel)

(* ******************************************************************
    function: build_enum_elements
    
    Return a java-valid string which helps creating the enum
    and adding its element into a global enum
            
    Input: name of global enum used in java code [string]
           identifier [string],
           symbols [StingHash],
           nesting level [int]
    Output: [string]
    **************************************************************** *)                    
let rec build_enum_elements glenum idstr lst symbols nestingLevel= match lst with        
    [] -> ""
|   Identifier(el)::[] -> glenum ^ ".add(\"" ^ idstr ^ "." ^ 
                          el ^ "\",\"" ^ idstr ^ "\");"      
|   Identifier(el)::l  -> glenum ^ ".add(\"" ^ idstr ^ "." ^ 
                          el ^ "\",\"" ^ idstr ^ "\");\n" ^
                          (build_enum_elements glenum idstr l symbols nestingLevel)  
(* ******************************************************************
    function: build_nested_collection
    
    Return a java-valid string which helps creating 
    the nested collection
            
    Input: a list of collection elemtns [Ast.expression],
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash],
           nesting level [int]
    Output: [string]
    **************************************************************** *)  
let rec build_nested_collection lst currentFunction funcArgs symbols nestingLevel= match lst with
    [] -> ""
|   Collection(el)::[] -> "new cCollection(new Object[]{"^
                          (build_nested_collection el currentFunction funcArgs symbols nestingLevel)^
                          "})"
|   Collection(el)::l ->  "new cCollection(new Object[]{"^
                          (build_nested_collection el currentFunction funcArgs symbols nestingLevel)^
                          "})," ^
                          build_nested_collection l currentFunction funcArgs symbols nestingLevel 
|   el::[] -> (eval_expr el currentFunction funcArgs symbols nestingLevel)
              
|   el::l -> (eval_expr el currentFunction funcArgs symbols nestingLevel)^
             "," ^
             build_nested_collection l currentFunction funcArgs symbols nestingLevel 
(* ******************************************************************
    function: build_nested_collection
    
    Return a java-valid string which helps creating 
    the a collection
            
    Input: identifier [string],
           a list of collection elemtns [Ast.expression],
           current function [string],
           a list of [Ast.javaType] which 
           represents the type of function arguments,
           symbols [StingHash],
           nesting level [int]
    Output: [string]
    **************************************************************** *) 
let rec build_collection_elements idstr lst currentFunction funcArgs symbols nestingLevel= match lst with
    [] -> ""
|   Collection(el)::[] -> (String.make nestingLevel '\t')^ 
                          idstr ^ 
                          ".append(new cCollection(new Object[]{"^
                          (build_nested_collection el currentFunction funcArgs symbols nestingLevel)^
                          "}));"
|   Collection(el)::l ->  (String.make nestingLevel '\t')^ 
                          idstr ^ 
                          ".append(new cCollection(new Object[]{"^
                          (build_nested_collection el currentFunction funcArgs symbols nestingLevel)^
                          "}));\n" ^
                          build_collection_elements idstr l currentFunction funcArgs symbols nestingLevel
|   el::[]            ->  (String.make nestingLevel '\t')^ 
                          idstr ^ 
                          ".append(" ^ 
                          (eval_expr el currentFunction funcArgs symbols nestingLevel) ^ 
                          ");"
|   el::l             ->  (String.make nestingLevel '\t')^
                          idstr ^ 
                          ".append(" ^ 
                          (eval_expr el currentFunction funcArgs symbols nestingLevel) ^ 
                          ");\n"^
                          build_collection_elements idstr l currentFunction funcArgs symbols nestingLevel
                          
(* ******************************************************************
    function: cast_bool_string, cast_int_string,
              cast_float_string, cast_string_string
    
    Return a java-valid string to
    cast bool (int/ float/ string) value into other types
    If the type of output is invalid, throw an error
            
    Input: value [string],
           output type [string]
    Output: [string]
    **************************************************************** *) 
let cast_bool_string value outputtype = match outputtype with
    "boolean"   ->    value
|   "int"       ->    "("^value^"?1:0)"
|   "double"     ->    "("^value^"?1.0:0.0)"
|   "String"    ->    "("^value^"?\"1\":\"0\")"
|   _           -> raise (Failure ("Invalid cast from bool to " ^ outputtype))

let cast_int_string value outputtype = match outputtype with
    "boolean"   -> "("^value^"==0?true:false)"
    (* TODO: deal with negative int *)
|   "int"       -> value
|   "double"     -> value^".0"
|   "String"    -> "\""^value^"\""
|   _           -> raise (Failure ("Invalid cast from int to " ^ outputtype))

let cast_float_string value outputtype = match outputtype with
    "boolean"   -> "("^value^"==0?true:false)"
        (* TODO: deal with negative float *)
|   "int"       -> "((int)"^value^")"
|   "double"     -> value
|   "String"    -> "\""^value^"\""
|   _           -> raise (Failure ("Invalid cast from float to " ^ outputtype))

let cast_string_string value outputtype = match outputtype with
    "boolean"   -> "("^value^".length()>0?true:false)"
|   "int"       -> "Integer.parseInt("^value^")"
|   "double"     -> "Float.parseFloat("^value^")"
|   "String"    -> value
|   _           -> raise (Failure ("Invalid cast from string to " ^ outputtype))

let build_cast_string value inputtype outputtype expr symbols nestingLevel = match inputtype with
    JavaBool    -> (cast_bool_string value outputtype)
|   JavaInt     -> (cast_int_string value outputtype)
|   JavaFloat   -> (cast_float_string value outputtype)
|   JavaString  -> (cast_string_string value outputtype)
|   _           -> raise (Failure ("Invalid cast from " ^ (string_of_jtype inputtype) ^ " to " ^ outputtype))

(* ******************************************************************
    function: build_input
    
    Return a java-valid string to call input function in java
    dealing with different return type
    If the type of identifier which will accept the input value
    is invalid, throw an error
            
    Input: identifier [string],
           symbols [StringHash]
    Output: [string]
    **************************************************************** *) 
let build_input idstr symbols = 
    let inputtype = getVarType idstr symbols in
    match inputtype with
    JavaInt      -> "input.cInputInt()"
|   JavaFloat    -> "input.cInputFloat()"
|   JavaBool     -> "input.cInputBool()"
|   JavaString   -> "input.cInputString()"
|   _ -> raise (Failure "Invalid input type")

(* ******************************************************************
    function: string_of_assignment
   
    Get result of assignment as a java-valid string

    Input: identifier name [string], expression [Ast.javaType], 
           function name [string],
           a list of [Ast.javaType] represents the types of arguments,
           symbol table [StringHash]
           nesting level [int]
    Output: [string]
    **************************************************************** *)
let cast_collection_string t x y currentFunction funcArgs symbols nestingLevel= match t with
    "int" -> "Integer.parseInt("^
             x ^".find("^(eval_expr y  currentFunction funcArgs symbols nestingLevel)^").toString());"
|   "boolean" -> "Boolean.parseBoolean("^
                 x ^".find("^(eval_expr y  currentFunction funcArgs symbols nestingLevel)^").toString());"
|   "String"  -> x ^".find("^(eval_expr y  currentFunction funcArgs symbols nestingLevel)^").toString();"
|   "double"  -> "Double.parseDouble("^
                 x ^".find("^(eval_expr y  currentFunction funcArgs symbols nestingLevel)^").toString());"
|   _         -> x ^".find("^(eval_expr y  currentFunction funcArgs symbols nestingLevel)^");"   
    
let string_of_assignment idstr expr currentFunction funcArgs symbols nestingLevel= let str = (match expr with
    Cast(x,y) -> let outputtype = (string_of_jtype (getVarType idstr symbols)) in
                               let inputtype = (type_of_expr x currentFunction funcArgs symbols nestingLevel) in
                               let value = (eval_expr x  currentFunction funcArgs symbols nestingLevel) in
                               outputtype^" "^idstr^"="^(build_cast_string value inputtype outputtype expr symbols nestingLevel)^";"
|   Collection(x) -> let decl = "cCollection" ^ " " ^ idstr ^"="^"new cCollection();\n" in
                                       let lst = get_collection_as_list expr in
                                       decl ^ (build_collection_elements idstr lst currentFunction funcArgs symbols nestingLevel)
|   CollectionBounds(x,y) -> string_of_jtype (getVarType idstr symbols) ^ " " ^
                             idstr ^ "=" ^
                             "new cCollection("^(eval_expr x  currentFunction funcArgs symbols nestingLevel)^","^
                             (eval_expr y  currentFunction funcArgs symbols nestingLevel)^");"
|   CollectionIndex(Identifier(x), y) -> 
                             let t = string_of_jtype (getVarType idstr symbols) in 
                             t ^" "^ 
                             idstr ^ "=" ^(cast_collection_string t  x y currentFunction funcArgs symbols nestingLevel)                                                           
|   Input(s) -> string_of_jtype (getVarType idstr symbols) ^ " " ^ idstr^"="^(build_input idstr symbols) ^ ";"
|   Enum(lst)  -> let glenum = "globalEnum" in
                build_enum_elements glenum idstr lst symbols nestingLevel
|   Struct(lst) -> "new cStruct();"^ 
                 (eval_structdef_list lst currentFunction funcArgs symbols nestingLevel)
| _ -> string_of_jtype (getVarType idstr symbols) ^ " " ^ idstr^"="^(eval_expr expr currentFunction funcArgs symbols nestingLevel) ^ ";") in (* TODO: uncomment for tabbing (String.make nestingLevel '\t')^ *)str

(* ******************************************************************
    function: add_cast_result
   
    Add cast result into symbol table

    Input: expression [Ast.expression], 
           identifier [string], 
           function name [string],
           a list of [Ast.javaType] represents the types of arguments,
           symbol table [StringHash]
           nesting level [int]
    Output: [unit]
    **************************************************************** *)
let add_cast_result = fun el id currentFunction funcArgs symbols nestingLevel->
    ignore(addVar id (type_of_expr el currentFunction funcArgs symbols nestingLevel) symbols nestingLevel)

(* ******************************************************************
    function: save_arg_to_hash
   
    Add argument into symboltable, lable it as Indeterminate type
    And print symbol table

    Input: identifier [string], 
           symbol table [StringHash]
           nesting level [int]
    Output: [unit]
    **************************************************************** *)
let save_arg_to_hash idstr symbols nestingLevel = 
    ignore (print_endline ("Assigning to " ^ idstr));
    ignore (addVar idstr Indeterminate symbols nestingLevel);
    print_hash symbols nestingLevel
    
(* ******************************************************************
    function: save_arglist_to_hash
   
    Add a list of arguments into symboltable
    And print symbol table

    Input: identifier [string], 
           symbol table [StringHash]
           nesting level [int]
    Output: [unit]
    **************************************************************** *)
let rec save_arglist_to_hash arglist symbols nestingLevel = 
    match arglist with
    []                  -> ()
|   Identifier(a)::[]   -> save_arg_to_hash a symbols nestingLevel
|   Identifier(a)::b    -> ignore(save_arg_to_hash a symbols nestingLevel); save_arglist_to_hash b symbols nestingLevel
    
let eval_update a currentFunction funcArgs symbols nestingLevel =  match a with
    PlusUp(id, exp)     ->
        (* let t = type_of_expr exp symbols in *)
    let str = string_of_identifier id in
    (match exp with
            Integer(v) ->  (ignore(save_assignment_to_hash str (Integer(v + 1)) currentFunction funcArgs symbols nestingLevel); str ^ "+=" ^ (eval_expr exp currentFunction funcArgs symbols nestingLevel))
        |   Floating(v) -> (ignore(save_assignment_to_hash str (Floating(v +. 1.0)) currentFunction funcArgs symbols nestingLevel); str ^ "+=" ^ (eval_expr exp currentFunction funcArgs symbols nestingLevel))
        |   StringString(v) -> (ignore (save_assignment_to_hash str (StringString("concatenation not implemented")) currentFunction funcArgs symbols nestingLevel); str ^ "+=" ^ (eval_expr exp currentFunction funcArgs symbols nestingLevel))
        |   _ -> raise (Failure "Invalid +=")
    )
|   MinusUp(id, exp)     ->(
(*         let t = type_of_expr exp symbols in *)
    let str = string_of_identifier id in
    (match exp with
            Integer(v) ->  (ignore(save_assignment_to_hash str (Integer(v - 1)) currentFunction funcArgs symbols nestingLevel); str ^ "-=" ^ (eval_expr exp currentFunction funcArgs symbols nestingLevel))
        |   Floating(v) -> (ignore(save_assignment_to_hash str (Floating(v -. 1.0)) currentFunction funcArgs symbols nestingLevel); str ^ "-=" ^ (eval_expr exp  currentFunction funcArgs symbols nestingLevel))
        |   _ -> raise (Failure "Invalid -=")))
|   TimesUp(id, exp)     ->(
(*         let t = type_of_expr exp symbols in *)
    let str = string_of_identifier id in
    (match exp with
            Integer(v) ->  (ignore(save_assignment_to_hash str (Integer(v + 1)) currentFunction funcArgs symbols nestingLevel); str ^ "*=" ^ (eval_expr exp currentFunction funcArgs symbols nestingLevel))
        |   Floating(v) -> (ignore(save_assignment_to_hash str (Floating(v +. 1.0)) currentFunction funcArgs symbols nestingLevel); str ^ "*=" ^ (eval_expr exp  currentFunction funcArgs symbols nestingLevel))
        |   _ -> raise (Failure "Invalid *=")
    ))
|   DivideUp(id, exp)     ->
(*         let t = type_of_expr exp symbols in *)
    let str = string_of_identifier id in
    (match exp with
            Integer(v) ->  (ignore(save_assignment_to_hash str (Integer(v + 1)) currentFunction funcArgs  symbols nestingLevel); str ^ "/=" ^ (eval_expr exp currentFunction funcArgs symbols nestingLevel))
        |   Floating(v) -> (ignore(save_assignment_to_hash str (Floating(v +. 1.0)) currentFunction funcArgs symbols nestingLevel); str ^ "/=" ^ (eval_expr exp  currentFunction funcArgs symbols nestingLevel)
            )
        |   _ -> raise (Failure "Invalid /=")

    )
    
let make_indeterminate_list arglist = 
    List.map (fun f -> Indeterminate) arglist

let rec eval_assignment tree currentFunction funcArgs symbols nestingLevel = match tree with
    Assignment(Identifier(idstr), expr) -> 
        ignore (print_endline ("Assigning to " ^ idstr));
        ignore (save_assignment_to_hash idstr expr currentFunction funcArgs symbols nestingLevel);
        string_of_assignment idstr expr currentFunction funcArgs symbols nestingLevel
|   FunctionAssign(FunctionIdentifier(id, arglist), Block(b)) -> 
        ignore (addFunc id (make_indeterminate_list arglist) JavaVoid symbols);
        ignore (save_arglist_to_hash arglist symbols (nestingLevel+1));
        if (id = "PLAY") then ("public static void main(String[] args){"^(eval_block b currentFunction funcArgs symbols nestingLevel)^"\n}\n") else ("private static "^(string_of_jtype (getFuncType id symbols))^" "^id^"("^(eval_arglist arglist)^"){"^(eval_block b currentFunction funcArgs symbols nestingLevel)^"\n}\n")
|   Empty -> ""
(* 

and eval_assignment_ lst currentFunction funcArgs symbols nestingLevel = match lst with
    []      ->  ()
|   a::[]   ->  save_assignment_to_hash a currentFunction funcArgs symbols nestingLevel
|   a::b    ->  ignore(save_assignment_to_hash a currentFunction funcArgs
                    symbols nestingLevel);
                save_assignments_to_hash b currentFunction funcArgs symbols nestingLevel
 *)
and eval_statement tree currentFunction funcArgs symbols nestingLevel = let stmt = (match tree with
    Expression(a) -> eval_expr a currentFunction funcArgs symbols (nestingLevel)
|   LocalVar(a) -> eval_assignment a currentFunction funcArgs symbols nestingLevel
|   Update(a) ->  eval_update a currentFunction funcArgs symbols nestingLevel ^ ";"
|   Output(a) -> "System.out.print("^eval_expr a currentFunction funcArgs symbols nestingLevel^");"
|   Return(a) -> ignore(setFuncType currentFunction (type_of_expr a currentFunction funcArgs symbols nestingLevel) symbols );
    "return " ^ (eval_expr a currentFunction funcArgs symbols nestingLevel) ^ ";"
|   IfStructure(a) -> eval_if a currentFunction funcArgs symbols (nestingLevel)
|   WhileStructure(a, Block(b)) -> "while "^ (eval_expr a currentFunction funcArgs symbols (nestingLevel+1)) ^" {" ^ (eval_block b currentFunction funcArgs symbols nestingLevel)^"}"
|   RuleStructure(a, b) -> let assignments = (eval_assignment_list a 
        currentFunction funcArgs symbols (nestingLevel+1)) in "{\n"^(String.make (nestingLevel+1) '\t')^assignments^
        "\n"^(eval_rules_block b currentFunction funcArgs symbols
            (nestingLevel+1))^"\n"^(String.make nestingLevel '\t')^"}"
|   ForStructure(Identifier(a), b, Block(c)) -> "for(int i=0;i<"^
            (eval_expr b currentFunction funcArgs symbols (nestingLevel+1))^
            ".size();i++){\n" ^ (String.make (nestingLevel+1) '\t')^
            (* TODO: check type of local variable *)
            "int " ^ a ^ "=" ^ 
            (eval_expr b currentFunction funcArgs symbols (nestingLevel+1))^
            ".find((int)i);" ^
            (eval_block c currentFunction funcArgs symbols (nestingLevel))^
            "\n"^(String.make nestingLevel '\t')^"}")

            in (String.make nestingLevel '\t')^ stmt

(* eval_block does NOT include the curly braces. When you call eval_block, add the curly braces to the result of this function. *)
and eval_block tree currentFunction funcArgs symbols nestingLevel = match tree with
    [] -> ""
|   a::[] -> "\n" ^ (eval_statement a currentFunction funcArgs symbols (nestingLevel+1))
|   a::b  -> let bl = (eval_block b currentFunction funcArgs symbols nestingLevel) in 
    (bl ^"\n" ^(eval_statement a currentFunction funcArgs symbols (nestingLevel+1)))

and eval_rules_block tree currentFunction funcArgs symbols nestingLevel = match tree with 
      []    -> ""
|   RuleBlock(e,Block(b))::[]   -> (String.make nestingLevel '\t')^"if (" ^
             (eval_expr e currentFunction funcArgs  symbols nestingLevel) ^
             ") {" ^
               (eval_block b currentFunction funcArgs
                symbols nestingLevel) ^"\n"^(String.make nestingLevel '\t')^"}"
|   RuleBlock(e,Block(b))::l -> let block = (eval_block b currentFunction funcArgs symbols nestingLevel) in  (String.make nestingLevel '\t')^"if (" ^ 
             (eval_expr e currentFunction funcArgs symbols nestingLevel) ^ 
             ") {" ^ block ^ "\n"^(String.make nestingLevel '\t')^"}\n" ^
               (eval_rules_block l currentFunction funcArgs
                symbols nestingLevel)
                                                               
and eval_assignment_list lst currentFunction funcArgs symbols nestingLevel = match lst with 
    []      -> ""
|   el::[]  -> (eval_assignment el currentFunction funcArgs symbols nestingLevel)
|   el::l   -> let first = (eval_assignment el currentFunction funcArgs symbols nestingLevel) in first^"\n"^(String.make nestingLevel '\t')^
               (eval_assignment_list l currentFunction funcArgs symbols nestingLevel)

and string_of_ifstatement kw pred bl currentFunction funcArgs symbols nestingLevel = match kw with
    "else"  -> kw ^ "{" ^ (eval_block bl currentFunction funcArgs symbols (nestingLevel)) ^ "\n}"
|   "elseif"-> "else if" ^ 
                "(" ^
                (eval_expr pred currentFunction funcArgs symbols nestingLevel) ^ 
                ")" ^(String.make (nestingLevel+1) '\t')^"{" ^ 
                (eval_block bl currentFunction funcArgs symbols (nestingLevel)) ^ "\n}"
|   _       -> kw ^
               "(" ^
               (eval_expr pred currentFunction funcArgs symbols nestingLevel) ^
               ") {" ^ 
                (eval_block bl currentFunction funcArgs symbols (nestingLevel)) ^ 
               "\n"^(String.make (nestingLevel) '\t')^"}"

and eval_if ifst currentFunction funcArgs symbols nestingLevel = match ifst with
    (* kw is the keyword if/else if/else
       pred is the predicate, just an expression that must have boolean type
       bl is the body block 
    *)
    [] -> ""
|    IfStatement(kw, pred, Block(bl))::[] -> string_of_ifstatement kw pred bl currentFunction funcArgs symbols (nestingLevel)
|    IfStatement(kw, pred, Block(bl))::b -> kw ^ 
               " ("^
               (eval_expr pred currentFunction funcArgs symbols (nestingLevel)) ^ 
               ")"^(String.make (nestingLevel+1) '\t')^ "{" ^
               (eval_block bl currentFunction funcArgs symbols (nestingLevel))^ 
               "\n"^(String.make (nestingLevel+1) '\t')^"}\n" ^
               (eval_if b currentFunction funcArgs symbols (nestingLevel)) 
let eval_top_level_item a symbols nestingLevel = 
match a with
    Assignment(_, _) -> eval_assignment a "" [] symbols nestingLevel
|   FunctionAssign(FunctionIdentifier(name,args),_) -> let fdef = eval_assignment a name args symbols nestingLevel in ignore (clearScope (nestingLevel+1) symbols); fdef
|   Empty -> ""


let rec eval_program tree symbols nestingLevel = match tree with
    (* Evaluate the program *)
    (* tree is an Ast object, symbols is a StringHash, nestingLevel is an int *) 
    [] -> ""
|   a::[] -> eval_top_level_item a symbols nestingLevel
|   a::b -> (
        let rest = (eval_program b symbols nestingLevel) in
        rest ^"\n"^(eval_top_level_item a symbols nestingLevel)
    )
                                                    
(* ******************************************************************

    Read in source code and perform lexical and syntactic analysis
    Find class and output file name
    
    ****************************************************************** *)
                                                    
let rec hyphens_to_camelcase str = match (String.contains str '-') with
    false   -> str
|   true    -> (
    let ind = String.index str '-' in 
    let firstpart = String.sub str 0 (ind) in
    let secondpart = String.sub str (ind+2) ((String.length str)-ind-2) in
    let ch = Char.uppercase (String.get str (ind+1))in
    hyphens_to_camelcase (firstpart ^ (String.make 1 ch) ^ secondpart))

let get_class_name = fun filename ->
    (* Determine the name for the java class from the filename *)
    let startIdx = String.rindex filename '/' + 1 in 
    let length = String.length filename - startIdx in
    let shortname = String.sub filename startIdx length in
    let capitalized = String.capitalize shortname in
    hyphens_to_camelcase capitalized

let get_java_file_name = fun filename -> 
    let startIdx = String.rindex filename '/' + 1 in
    let lower = String.get filename startIdx in
    let upper = Char.uppercase lower in
    ignore (String.set filename startIdx upper);hyphens_to_camelcase filename

let compile = fun source class_name ->
    (* Compile function inserts compiled code into a java class *)
    let lexbuf = Lexing.from_string source in
    let tree = (Parser.program Scanner.token lexbuf) in
    let symbols = StringHash.create 5 in
    let result = eval_program tree symbols 0 in
    sprintf "import java.io.BufferedReader;\nimport java.io.IOException;\nimport java.io.InputStream;\nimport java.io.InputStreamReader;\nimport java.util.Scanner;\n\npublic class %s{\nstatic cInput input = new cInput();\n//Begin Cardigan Code\n\n%s\n//End Cardigan Code\n}\n" (get_class_name class_name) result 
;;

let get_source = fun filename ->
    (* Reads in the cardigan source file as a string *)
    (* This is sutable to pass to Lexing.from_string *)
    let source_file = open_in filename in 
    let source_len = in_channel_length source_file in
    let source = String.create source_len in
    really_input source_file source 0 source_len;
    close_in source_file;
    
    (* Trim trailing whitespace. *)
    (* We may need to leave this in once the interpreter is hooked up *)
    String.trim source
;;

let strip_extension = fun full ->
    (* Returns the filename with the extension stripped *)
    let idx = String.index full '.' in
    String.sub full 0 idx
;;

let _ = 
    (* Determine the filename without extension *) 
    let filename = strip_extension Sys.argv.(1) in 

    (* Read in the cardigan source code *)
    let cardigan_source = get_source (filename ^ ".crd") in
    
    (* Compile the cardigan source into java *)
    let java_source = compile cardigan_source filename in
    
    (* Write the compiled code to a .java file *)
    let target = open_out ((get_java_file_name filename) ^ ".java") in 
    output_string target java_source; close_out target
;;

