%{open Ast %}

%token<string> ID

/* Primitive types */
%token<int> INT
%token<bool> BOOLEAN
%token<float> FLOAT
%token<string> STRING

/* Grouping symbols */
%token LPAREN RPAREN RCURL LCURL RBRAC LBRAC BAR

/* Punctuation */
%token COMMA DOT COLON ASSIGN

/* Arithmetic Operators */
%token PLUS MINUS TIMES DIVIDE MOD

/* Boolean Operators */
%token NOT AND OR

/* Comparison Operators */
%token NEQ EQ LT GT LEQ GEQ

/* Update Operators */
%token PLUSEQUALS MINUSEQUALS TIMESEQUALS DIVIDEEQUALS

/* Separators */
%token EOL EOF

/* Keywords */
%token IF ELSEIF ELSE WHILE FOR RETURN RULES

/* Built in functions */
%token CAST PLAY INPUT OUTPUT

/* Type keywords */
%token INTTYPE FLOATTYPE STRINGTYPE BOOLTYPE

/* Increment/Decrement keywords */
%token INCREMENT DECREMENT

/* Associativity rules */
%right ASSIGN
%left EQ NEQ LT GT LEQ GEQ
%left AND OR
%left PLUS MINUS
%left TIMES DIVIDE MOD
%left NOT
%left INCREMENT DECREMENT DOT

/* Start info */
%start program
%type <Ast.program> program
%%

arguments:
	/* no args */									{[]}	
|	ID												{[Identifier($1)]}
|	ID COMMA arguments								{Identifier($1)::$3}

identifier:
   ID 												{Identifier($1)}

functionId:
	ID LPAREN arguments RPAREN						{FunctionIdentifier($1, $3)}

block:
	LCURL statementList RCURL						{Block($2)}

ruleBlock:
	expression block								{[RuleBlock($1, $2)]}
|	expression block ruleBlock						{RuleBlock($1, $2)::$3}

expressionList:
	/* empty list */								{[]}
|	expression										{[$1]}
|	expression COMMA expressionList 				{$1::$3}

structDef:
	ID COLON expression								{StructDef($1, $3)}

structBody:
	/* empty struct */								{[]}
|	structDef										{[$1]}
|	structDef COMMA structBody						{$1::$3}

typeKeyword:
	INTTYPE											{"int"}
|	FLOATTYPE										{"float"}
|	STRINGTYPE										{"String"}
|	BOOLTYPE										{"boolean"}

binaryOperation:
	expression PLUS expression						{Plus($1,$3)}
|	expression MINUS expression 					{Minus($1,$3)}
|	expression TIMES expression 					{Times($1,$3)}
|	expression DIVIDE expression 					{Divide($1,$3)}
|	expression MOD expression						{Mod($1,$3)}
|	expression AND expression 						{And($1,$3)}
|	expression OR expression						{Or($1,$3)}
|	expression NEQ expression 						{Neq($1,$3)}
|	expression EQ expression 						{Eq($1,$3)}
|	expression LT expression 						{Lt($1,$3)}
| 	expression GT expression 						{Gt($1,$3)}
|	expression LEQ expression 						{Leq($1,$3)}
|	expression GEQ expression 						{Geq($1,$3)}

expression:
 	INT												{Integer($1)}
|	BOOLEAN 										{Boolean($1)}
|	FLOAT 											{Floating($1)}
|	STRING 											{StringString($1)}
|	identifier 										{IdExpr($1)}
|	LBRAC expression COLON expression RBRAC 		{CollectionBounds($2, $4)}
|   LBRAC expressionList RBRAC 						{Collection($2)}
|	BAR arguments BAR								{Enum($2)}
|	LPAREN expression RPAREN						{ParenExpr($2)}
|	LCURL structBody RCURL							{Struct($2)}
|	CAST LPAREN expression COMMA typeKeyword RPAREN	{Cast($3, $5)}
|	INPUT LPAREN typeKeyword RPAREN					{Input($3)}
|	MINUS expression								{Negate($2)}
| 	NOT expression									{Invert($2)}
|	binaryOperation									{BinaryOp($1)}
|	expression INCREMENT							{Increment($1)}
|	expression DECREMENT							{Decrement($1)}
|	identifier LBRAC expression RBRAC				{CollectionIndex($1, $3)}
|	identifier DOT expression						{Lookup($1, $3)}
|	functionId										{FunctionCall($1)}

statementList:
	statement 										{[$1]}
|	statementList EOL								{$1}
|	statementList EOL statement 					{$3::$1}

idUpdate:
	identifier PLUSEQUALS expression				{PlusUp($1, $3)}
|	identifier MINUSEQUALS expression				{MinusUp($1, $3)}
|	identifier TIMESEQUALS expression 				{TimesUp($1, $3)}
|	identifier DIVIDEEQUALS expression 				{DivideUp($1, $3)}	
    
elseStatements:
	ELSE block										{[IfStatement("else", Boolean(true), $2)]}
|	ELSEIF expression block							{[IfStatement("else if", $2, $3)]}
|	ELSEIF expression block elseStatements			{IfStatement("else if", $2, $3)::$4}

assignmentList:
	/*empty*/ 										{[]}
|	assignment 										{[$1]}	
|	assignment COMMA assignmentList					{$1::$3}

statement:
	expression										{Expression($1)}
|	assignment										{LocalVar($1)}
|	idUpdate										{Update($1)}
|	OUTPUT LPAREN expression RPAREN 				{Output($3)}
|	RETURN expression								{Return($2)}
|	IF expression block								{IfStructure(IfStatement("if", $2, $3)::[])}
|	IF expression block elseStatements				{IfStructure(IfStatement("if", $2, $3)::$4)}
|	WHILE expression block							{WhileStructure($2, $3)}
|	RULES LCURL ruleBlock RCURL						{RuleStructure([], $3)}
|	RULES LPAREN assignmentList RPAREN LCURL ruleBlock RCURL	{RuleStructure($3, $6)}
|	FOR identifier COLON expression block			{ForStructure($2, $4, $5)}

assignment:
	identifier ASSIGN expression 					{Assignment($1, $3)}
|	functionId ASSIGN block 						{FunctionAssign($1, $3)}

program:
	assignment 										{[$1]}
|	program EOL assignment 							{($3::$1)}
